########################################################################
# Copyright (C) 2010 VMWare, Inc.
# All Rights Reserved
########################################################################

"""This module contains the HostImage class,
   used for applying and extracting an image
   profile from the host.
"""

import datetime
import os
import logging
from operator import attrgetter

import Installer
import Installer.BootBankInstaller
import Installer.LiveImageInstaller
import Installer.LockerInstaller
import Database
import Errors
import Downloader
import Vib
import VibCollection
from vmware import runcommand
import vmware.vob
from Utils import HashedStream, LockFile, EsxGzip
from Utils.XmlUtils import _utctzinfo

try:
   import spicy
   HAVE_SPICY = True
except ImportError:
   HAVE_SPICY = False


DEFAULT_INSTALLERS = (Installer.BootBankInstaller.BootBankInstaller,
                      Installer.LiveImageInstaller.LiveImageInstaller,
                      Installer.LockerInstaller.LockerInstaller,)
INSTALLERTYPE_TO_NAME = {'boot' : 'BootBankInstaller',
      'live' : 'LiveImageInstaller',
      'locker': 'LockerInstaller'}

ADVCFG = '/sbin/esxcfg-advcfg'
VMKVOB = '/usr/lib/vmware/vob/bin/vmkvob'
LOCKFILE = '/var/run/esximg.pid'

#
# This is the default acceptance level that gets returned if the configuration
# cannot be found.
DEFAULT_HOST_ACCEPTANCE = Vib.ArFileVib.ACCEPTANCE_PARTNER

log = logging.getLogger('HostImage')


class HostImage(object):
   """The HostImage class enables the basic operation of applying
      an image profile to an ESX host via the Stage and Remediate
      methods.  It also provides a way to obtain the current
      VIB inventory and current image profile.

      Class Variables:
         * IMGSTATE_FRESH_BOOT - The host is local booted and VisorFS VIBs
                        are the same as /bootbank VIBs (no VIBs have been
                        updated); or the host is PXE booted
         * IMGSTATE_LIVE_UPDATED - VIBs have been live installed or removed
                        but no reboot-required VIBs have been installed
                        since the last boot.  The VIBs in VisorFS are diff-
                        erent from the /bootbank VIBs, but should be the same
                        as /altbootbank VIBs.
         * IMGSTATE_BOOTBANK_UPDATED - VIBs requiring a reboot have been
                        installed, so that the VIBs in /altbootbank are diff-
                        erent than the VIBs in VisorFS.  Once a host enters
                        this state, live installs are not permitted until
                        after a reboot.

      Attributes:
         * installers    - A dict of supported Installer classes.
         * imgstate      - Current ESX image state, one of the IMGSTATE_* values
         * livevibs      - A VibCollection instance of live VIBs
         * livestagedvibs - A VibCollection instance of VIBs staged for
                            LiveImageInstaller.
         * bootbankupdatedvibs - A VibCollection of VIBs, which are installed
                                 into altbootbank and will be live after reboot.
         * bootbankstagedvibs  - A VibCollection of VIBs, which are in
                                 altbootbank, but altbootbank is in staged
                                 state.
   """
   IMGSTATE_UNKNOWN = -1
   IMGSTATE_FRESH_BOOT = 0
   IMGSTATE_LIVE_UPDATED = 1
   IMGSTATE_BOOTBANK_UPDATED = 2

   DB_AUTO = 0
   DB_VISORFS = 1
   DB_BOOTBANK = 2
   DB_CURBOOTBANK = 3
   DBS = (DB_AUTO, DB_VISORFS, DB_BOOTBANK, DB_CURBOOTBANK)

   # Use the same buffer size as urlgrab...
   # if the size is too small then slow links will hang...
   BUFFER_SIZE = 8*1024

   def __init__(self, installerclasses = DEFAULT_INSTALLERS):
      """HostImage class constructor.
         Initialize all the installers.
         Note, on PXE systems the BootbankInstaller and LockerInstaller
         will not initialize and will not be present.
         Parameters:
            * installerclasses - a list or iterable of class objects
                                 inherited from Installer.Installer
         TODO: Exceptions:
            SetupError - if no installer classes could be initialized
      """
      self.installers = {}

      for cls in installerclasses:
         try:
            installer = cls()
            self.installers[installer.installertype] = installer
         except Exception, e:
            if hasattr(e, 'msg'):
               log.info('Installer %s was not initiated - reason: %s' %
                        (str(cls), e.msg))
            else:
               log.info('Installer %s was not initiated - reason: %s' %
                        (str(cls), str(e)))
      #
      # If bootbank installer is not availalbe, then we don't need to
      # handle tools package with locker installer.
      #
      if 'locker' in self.installers and \
         'boot' not in self.installers:
         log.info('BootbankInstaller is not initialized, no need to keep '
                  '%s' % self.installers['locker'].__class__.__name__)
         del self.installers['locker']

      self._lock = LockFile.LockFile(LOCKFILE)
      self._hosthw = None

   def _getImgState(self):
      if 'boot' not in self.installers:
         if 'live' not in self.installers:
            return self.IMGSTATE_UNKNOWN
         # Probably PXE
         # XXX: if we support fresh install, there might be a drift
         return self.IMGSTATE_FRESH_BOOT
      elif 'live' not in self.installers:
         # bootbank exists but not live installer, probably offline
         # State depends on whether /altbootbank image has been updated
         bootinst = self.installers['boot']
         if bootinst.bootbankstate == bootinst.BOOTBANK_STATE_UPDATED:
            return self.IMGSTATE_BOOTBANK_UPDATED
         else:
            return self.IMGSTATE_FRESH_BOOT
      else:
         # Both bootbank and live images exist.
         bootinst = self.installers['boot']
         liveinst = self.installers['live']
         #
         # Altbootbank contains a valid, newer image than bootbank.
         # If altbootbank image is the same as the live image, then
         # the live image has been updated as well. If they are different,
         # then altbootbank is newer than the live image.  It is not possible
         # for the live image to be newer than /altbootbank once /altbootbank
         # is updated, we prevent that scenario.
         if bootinst.bootbankstate == bootinst.BOOTBANK_STATE_UPDATED:
            if liveinst.database.vibIDs == bootinst.database.vibIDs:
               return self.IMGSTATE_LIVE_UPDATED
            else:
               return self.IMGSTATE_BOOTBANK_UPDATED
         # Altbootbank contains an older image than /bootbank.
         # Most likely this is a fresh boot, so the live image is the same as
         # the bootbank image.
         # It is possible that a live install happened, but for some reason the
         # altbootbank was not updated.  In this case, we still return live
         # updated, but if a user reboots, they will lose their live image changes.
         elif bootinst.database.vibIDs == liveinst.database.vibIDs:
            return self.IMGSTATE_FRESH_BOOT
         else:
            log.debug('Live image has been updated but /altbootbank image has '
                      'not.  This means a reboot is not safe.')
            return self.IMGSTATE_LIVE_UPDATED

   imgstate = property(_getImgState)

   def _getDatabase(self, database = DB_AUTO):
      if database not in self.DBS:
         raise ValueError("Database value of %s not valid" % (database))

      #
      # Pick database automatically.  Pick from bootbank database if
      # bootbank-only remediation has occured; otherwise from live database
      #
      if database == self.DB_AUTO:
         if self.imgstate == self.IMGSTATE_BOOTBANK_UPDATED:
            database = self.DB_BOOTBANK
         elif 'live' in self.installers:
            database = self.DB_VISORFS
         else:
            database = self.DB_CURBOOTBANK

      if database == self.DB_BOOTBANK and 'boot' in self.installers:
         return self.installers['boot'].database
      elif database == self.DB_VISORFS and 'live' in self.installers:
         return self.installers['live'].database
      elif database == self.DB_CURBOOTBANK and 'boot' in self.installers:
         return self.installers['boot'].bootbank.db
      else:
         return Database.Database('')

   def GetInventory(self, database = DB_AUTO):
      """Obtains the collection of VIBs installed on the ESX host.
         If the host is PXE-booted or in the 'Fresh Boot' and
         'Live Image Updated' states, then the VIB inventory
         will be taken from the VisorFS database.  If the host
         is in the 'Boot Image Updated' state, then the VIB
         inventory is taken from the /altbootbank database. If database is
         DB_AUTO, VIBs from locker will be added to the return value.

         Parameters:
            * database - Override the automatic database selection algorithm
                         and read the inventory from a specific database.
                         Should be one of the DB_* class variables.

         Returns:
            A VibCollection instance, or None if no database was found

         Exceptions:
            ValueError  - Illegal database value passed
            DatabaseFormatError - Database files not formatted correctly
            DatabaseIOError     - Database could not be read
      """
      allvibs = VibCollection.VibCollection()
      allvibs += self._getDatabase(database).vibs
      # locker VIBs are only kept in locker installer
      if 'locker' in self.installers:
         allvibs += self.installers['locker'].database.vibs
      return allvibs

   def GetProfile(self, database = DB_AUTO):
      """Obtains the last image profile applied to the ESX host.
         If the host is PXE-booted or in the 'Fresh Boot' and
         'Live Image Updated' states, then the image profile
         will be taken from the VisorFS database.  If the host
         is in the 'Boot Image Updated' state, then the image
         profile is taken from the /altbootbank database.
         Locker VIBs are added back to image profile because they are saved
         separately in product locker.

         Parameters:
            * database - Override the automatic database selection algorithm
                         and read the profile from a specific database.
                         Should be one of the DB_* class variables.

         Returns:
            An ImageProfile instance, or None if no image profile
            is in the database.

         Exceptions:
            ValueError  - Illegal database value passed
            DatabaseFormatError - Database files not formatted correctly
            DatabaseIOError     - Database could not be read
      """
      db = self._getDatabase(database)
      profile = None
      if db.profile:
         profile = db.profile.Copy()
         profile.creationtime = db.profile.creationtime
         profile._modifiedtime = db.profile._modifiedtime
      # Host acceptance level overrides image profile's
      # NOTE: Always retrieve the acceptance level each time
      # It may have changed in the background as there are multiple
      # methods of changing it

      # Please ignore the above comment. As a fix of bug 985333, it is noted that
      # profile acceptance level in the database never gets changed, so reseting it to
      # hostacceptance level is not right. Add to that, profile acceptancelevel is not printed by
      # 'esxcli software profile get' command and the below code modifies the _modifiedtime attribute
      # through the acceptancelevel property of the profile
      # which results in wrong modification time of the profile.

      #hostacceptance = self._getHostAcceptance()
      #if hostacceptance and profile:
      #   profile.acceptancelevel = hostacceptance

      # Adding back locker VIBs as live/boot installer doesn't have locker VIBs
      # Modify the profile._modifiedtime to the greater of the db.profile._modifiedtime
      # or the latest of the vib installdate in the locker database
      if profile and 'locker' in self.installers:
         lockervibs = self.installers['locker'].database.vibs
         prevprofilemodifiedtime = profile._modifiedtime
         for vib in lockervibs.values():
            profile.AddVib(vib)
            if prevprofilemodifiedtime < vib.installdate:
               prevprofilemodifiedtime = vib.installdate
         profile._modifiedtime = prevprofilemodifiedtime

      return profile

   @property
   def livevibs(self):
      vibs = VibCollection.VibCollection()
      if 'live' in self.installers:
         vibs.update(self.GetInventory(self.DB_VISORFS))
      return vibs

   @property
   def livestagedvibs(self):
      vibs = VibCollection.VibCollection()
      if 'live' in self.installers:
         livevibids = set(self.livevibs.keys())
         db = self.installers['live'].stagedatabase
         if db is not None:
            for k, v in db.vibs.iteritems():
               if k not in livevibids:
                  vibs.AddVib(v)
      return vibs

   @property
   def bootbankupdatedvibs(self):
      vibs = VibCollection.VibCollection()
      if self.imgstate == self.IMGSTATE_BOOTBANK_UPDATED:
         # Altbootbank has the latest ImageProfile, GetInventory will return
         # vibs from latest DB by default.
         vibs.update(self.GetInventory())
      return vibs

   @property
   def bootbankstagedvibs(self):
      vibs = VibCollection.VibCollection()
      if 'boot' in  self.installers:
         db = self.installers['boot'].stagedatabase
         if db is not None:
            vibs.update(db.vibs)
      return vibs

   def GetHostHwPlatform(self):
      if self._hosthw is None:
         from Utils import HostInfo
         vendor, model = HostInfo.GetBiosVendorModel()
         self._hosthw = Vib.HwPlatform(vendor, model)
      return self._hosthw

   def GetHostAcceptance(self):
      """Returns the host acceptance level setting.  If there is no setting
         or if the configured value is invalid, then a default acceptance
         level will be returned.
         Parameters:
            None
         Returns:
            One of the Vib.ArFileVib.ACCEPTANCE_LEVELS settings.
         Raises:
            ValueError -- Illegal value retrieved
      """
      return self._getHostAcceptance()

   def SetHostAcceptance(self, newlevel):
      """Sets the host acceptance level to a new level.  Validates that
         the new level still works with the current image profile, ie that no
         VIBs in the current image has a lower level than the new level.
         To prevent the acceptance level from changing while an image is
         being updated, locking will be used.
         Parameters:
            * newlevel - the new acceptance level, should be one of the constants
                         Vib.ArFileVib.ACCEPTANCE_LEVELS
         Returns:
            None
         Raises:
            LockError   - Image is being updated, it's not safe to change the level
            AcceptanceConfigError - new level invalidates the image profile, or
                                    unable to run esxcfg-advcfg
            ValueError  - "newlevel" is not a valid acceptance level
      """
      if newlevel not in Vib.ArFileVib.ACCEPTANCE_LEVELS:
         raise ValueError("Invalid acceptance level of '%s' was passed in" % (newlevel))

      self._getLock()
      try:
         prof = self.GetProfile()
         oldlevel = prof.acceptancelevel
         log.info("Attempting to change the host acceptance level from %s to %s"
                  % (oldlevel, newlevel))
         prof.acceptancelevel = newlevel
         prof.vibs = self.GetInventory()

         # Focus on only acceptance level problems
         # TODO: Disable extensibility rule checks
         problems = prof.Validate(nodeps=True, noconflicts=True,
                                  allowobsoletes=True,
                                  allowfileconflicts=True)
         if problems:
            badvibs = [getattr(p, 'vibid', '') for p in problems]
            msg = "Unable to set acceptance level of %s due to installed VIBs "\
                  "%s having a lower acceptance level." \
                  % (newlevel, ', '.join(badvibs))
            raise Errors.AcceptanceConfigError(msg)

         self._setHostAcceptance(newlevel)
         # See #bora/apps/addvob/addvob.c for the vob format string.
         self.SendVob("hostacceptance.changed", oldlevel, newlevel)
      finally:
         self._freeLock()

   def _GetOrderedInstallers(self):
      return sorted(self.installers.values(), key=attrgetter('priority'))

   def _download_and_stage(self, imgprofile, vibid, installer,
                           checkacceptance = True, schema = None):
      arvib = src = None
      vib = imgprofile.vibs[vibid]
      log.info('Attempting to download VIB %s' % (vib.name))
      for url in vib.remotelocations:
         try:
            d = Downloader.Downloader(url)
            src = d.Open()
            arvib = Vib.ArFileVib.FromFile(src, schema)
            break
         except Downloader.DownloaderError, e:
            if src:
               src.close()
               src = None
            log.info('Unable to download from %s, error [%s]. Trying next url...'
                     % (url, str(e)))
            continue
      if not arvib:
         raise Errors.VibDownloadError(vib, "Unable to download VIB "
                                       "from any of the URLs %s" % (
                                       ', '.join(vib.remotelocations)))

      # Verify descriptor matches metadata:
      try:
         arvib.MergeVib(vib)
      except (ValueError, TypeError), e:
         raise Errors.InstallationError(arvib.id, str(e))

      if checkacceptance:
         try:
            arvib.VerifyAcceptanceLevel()
         except (Errors.VibFormatError, Errors.VibValidationError,
                 Errors.VibSignatureError), e:
            raise Errors.InstallationError(arvib.id, str(e))

      # TODO:
      #   - verify hardware platform

      # Loop through each payload, open dest payload file, and start buffer
      # copy.  Compute digest.
      for payload, sourcefp in arvib.IterPayloads():
         if payload.name in imgprofile.vibstates[vibid].payloads:
            payload.localname = imgprofile.vibstates[vibid].payloads[payload.name]
         else:
            payload.localname = None

         try:
            dfp = installer.OpenPayloadFile(arvib.id, payload, write=True,
                  read=False)
            if not dfp:
               log.debug('Payload %s from vib %s skipped by %s' %
                         (payload.name, arvib.id, installer.__class__.__name__))
               continue
            if payload.checksums:
               checksum = payload.GetPreferredChecksum()
               hashalgo = checksum.checksumtype.replace("-", "")
               sourcefp = HashedStream.HashedStream(sourcefp,
                                                    checksum.checksum,
                                                    hashalgo)
            inbytes = sourcefp.read(self.BUFFER_SIZE)
            dfp.write(inbytes)
            while inbytes:
               inbytes = sourcefp.read(self.BUFFER_SIZE)
               dfp.write(inbytes)
            sourcefp.close()
            dfp.close()
         except Exception, e:
            installer.Cleanup()
            src.close()
            raise Errors.InstallationError(arvib.id, str(e))

      # Explicitly check for installdate is None. If installdate is already
      # set, we do not want to reset it.
      if vib.installdate is None:
         vib.installdate = datetime.datetime.now(_utctzinfo)

      # Close VIB
      src.close()

   def _get_tardisk_payload(self, imgprofile, vibid, vib):
      # Construct vib compressed payloads for /altbootbank
      # from uncompressed payloads in /tardisk.
      vibstate = imgprofile.vibstates[vibid]
      for pl in vib.payloads:
         if pl.payloadtype not in Installer.BootBankInstaller.\
            BootBankInstaller.SUPPORTED_PAYLOADS:
            continue

         plname = pl.name
         srcfile = None
         if plname in vibstate.payloads:
            srcname = vibstate.payloads[plname]
            srcfile = os.path.join('/tardisks', srcname)
            destfile = os.path.join('/altbootbank', srcname)

         if srcfile is not None and os.path.isfile(srcfile):
            # gzip the payload from uncompressed file
            src = EsxGzip.GunzipFile(srcfile, 'rb')
            dst = open(destfile, 'w')
            log.info('Constructing %s from %s' %(destfile, srcfile))

            filesize = 0
            while True:
               inbytes = src.read(self.BUFFER_SIZE)
               if not inbytes:
                  break
               dst.write(inbytes)
               filesize += len(inbytes)

            log.debug('Total of %d bytes were written.' % (filesize))
            src.close()
            dst.close()
         else:
            msg = 'Cannot locate source for payload %s of '\
                  'VIB %s ' % (plname, vibid)
            raise Errors.InstallationError(str(imgprofile.vibIDs), msg)

   def Stage(self, imgprofile, nosigcheck = False,
                   forcebootbank = False, dryrun = False,
                   vibschema = None, stageonly = False, checkacceptance = True):
      """Stage prepares or 'stages' a new ESX image for application
         based on the specified image profile.  If the current image
         or last staged image is composed of the same VIBs as the
         passed in image profile, then the Stage method will have
         no work to do.
         Maintenance Mode is not required during a stage operation.
         Stage also checks for any hwplatform dependencies.

         The VIBs are downloaded and a new ESXi image is prepared.  The
         StartTransaction, OpenPayloadFile, and CompleteStage methods of
         *Installer subclasses actually carry out the work.  This Stage method
         is responsible for coordinating between different Installers,
         VIB signature validation, and driving each Installer to install
         successive VIB payloads.  The databases will be locked to prevent
         another process from calling Stage() at the same time.

         Stage manages the sourcing of VIBs, either initiating downloads or
         sourcing them from previous Installers for chained installation.

         TBD: Cycling may not be practical.
         We want to skip LiveImageInstaller once the image state becomes
         BOOTBANK_UPDATED or forcebootbank is True.

         Parameters:
            * imgprofile   - The ImageProfile instance to stage. It must have
                             a populated vibs attribute.
            * nosigcheck   - Boolean, skips signature validation if True
            * forcebootbank - Force an install using BootbankInstaller even
                              if the LiveImageInstaller could be used
            * dryrun       - Dry run only, report what will be changed
                             but do not make changes to the system.
            * vibschema    - A file path to an XML schema for validation
                             of the VIB descriptor. Set to None to disable
                             validation.
            * checkacceptance - If True (the default), VIB acceptance levels
                                 will be validated. A failed validation raises
                                 an exception.

         Returns:
            None

         Exceptions:
            InstallationError
            LockingError      - Unable to lock database for writing. Most
                                likely another thread has invoked Stage().
            VibDownloadError  - error during download of a VIB
            VibFormatError    - improper VIB or descriptor structure
            VibSignatureError - invalid VIB signature
            DatabaseFormatError - Database files not formatted correctly
            DatabaseIOError     - Database could not be read
            NoMatchError      - VIB hwplatform dependency could not
                                be satisfied
      """
      assert len(imgprofile.vibs) >= len(imgprofile.vibIDs)
      log.debug('Staging image profile [%s]' % (imgprofile.name))
      log.debug('VIBs in image profile: %s' % (', '.join(imgprofile.vibIDs)))

      # Check potential installation problems before hand
      #   - current bootbank bootstate must be 0
      #   - installation size must fit in bootbank partition
      if 'boot' in self.installers:
         self.installers['boot'].PreInstCheck(imgprofile)

      # hwplatform requirement check
      hwmatchproblems = list()
      hosthw = self.GetHostHwPlatform()
      for imghw in imgprofile.GetHwPlatforms():
         prob = imghw.MatchProblem(hosthw)
         if prob is None:
            # We have a match, forget any other mismatches
            hwmatchproblems = list()
            break
         else:
            hwmatchproblems.append(prob)

      if len(hwmatchproblems) > 0:
         reasons = ["BIOS %s %s is required but host BIOS %s is %s" \
                    % (p[0], p[1], p[0], p[2]) for p in hwmatchproblems]
         msg = "Host doesn't meet image profile '%s' hardware requirements:" \
               "\n%s" % (imgprofile.name, '\n'.join(reasons))
         log.error(msg)
         # See #bora/apps/addvob/addvob.c for the vob format string.
         self.SendVob("install.invalidhardware", imgprofile.name, \
                         '\n'.join(reasons))
         raise Errors.NoMatchError(imgprofile.name, '\n'.join(reasons))

      # Send VOB warning if installing image profile with nosigcheck or
      # checkacceptance is False
      if nosigcheck or not checkacceptance:
         reasons = list()
         if nosigcheck:
            reasons.append("signature validation disabled")
         if not checkacceptance:
            reasons.append("acceptance level checking disabled")
         msg = "SECURITY ALERT: Installing image profile '%s' with %s." \
               % (imgprofile.name, " and ".join(reasons))
         log.warn(msg)
         # See #bora/apps/addvob/addvob.c for the vob format string.
         self.SendVob("install.securityalert", imgprofile.name, \
                         " and ".join(reasons))

      self._getLock()
      try:
         #
         # Cycle through each installer. Installers are ordered by its
         # priority. Current design is to run live installer first
         #
         installerstate = {'unsupported' : list(),
                           'nochange'    : list(),
                           'finished'    : list()}
         for installer in self._GetOrderedInstallers():
            # Skip bootbank stage if live staged
            installertype = installer.installertype
            if installertype == 'boot' and \
                  'live' in installerstate['finished']:
               log.debug("Already staged to live image, skipping "
                         "BootbankInstaller.")
               installerstate['finished'].append(installertype)
               continue

            try:
               adds, removes, staged = installer.StartTransaction(imgprofile,
                                 imgstate = self.imgstate,
                                 forcebootbank = forcebootbank,
                                 stageonly = stageonly,
                                 preparedest = not dryrun)
            except EnvironmentError, e:
               installer.Cleanup()
               raise Errors.InstallationError('', "Failed to start installation: "
                                              + str(e))

            # (None, NONE, False) is returned if installer could not handle the
            # transaction
            if adds is None and removes is None:
               log.info('%s is not supported, skipping.' %
                        (installer.__class__.__name__))
               installerstate['unsupported'].append(installertype)
               continue

            if not adds and not removes:
               log.info('Nothing for %s to do, skipping.' %
                        (installer.__class__.__name__))
               installerstate['nochange'].append(installertype)
               continue

            log.debug(' --- Stage: %s adding [%s], removing [%s]' % (
                      installer.__class__.__name__,
                      ', '.join(adds), ', '.join(removes)))

            if not (staged or dryrun):
               for vibid in adds:
                  vib = imgprofile.vibs[vibid]
                  if not vib.remotelocations:
                     # for vib installed after the host boots up,
                     # get vib payloads from /tardisks
                     log.info("Constructing payloads from tardisks for vib %s"
                               % vibid)
                     self._get_tardisk_payload(imgprofile, vibid, vib)
                     continue

                  # For now, always download VIBs that we need.
                  # This may be inefficient, but it will always work.
                  self._download_and_stage(imgprofile, vibid,
                                          installer, checkacceptance, vibschema)

               try:
                  installer.CompleteStage()
               except:
                  installer.Cleanup()
                  raise

            installerstate['finished'].append(installertype)

         if len(installerstate['unsupported']) == len(self.installers):
            msgs = list()
            for installer in self.installers.values():
               for problem in installer.problems:
                  msgs.append(str(problem))
            raise Errors.InstallationError('',
                  'The transaction is not supported: %s' % (' '.join(msgs)))

         if dryrun:
            installers = [INSTALLERTYPE_TO_NAME[inst] for inst in
                  installerstate['finished']]
            msg = ('Dryrun only, host not changed. '
                   'The following installers will be applied: [%s]' %
                   (', '.join(installers)))
            # If only boot bank installer is applied, then
            # live image is not changed - reboot is required
            if 'boot' in installerstate['finished'] and \
               'live' not in installerstate['finished']:
               raise Errors.NeedsRebootResult(msg)
            else:
               raise Errors.NormalExit(msg)
      except Errors.NormalExit:
         raise
      except Exception, e:
         msg = "Could not stage image profile '%s': %s" % (imgprofile.name, e)
         # See #bora/apps/addvob/addvob.c for the vob format string.
         self.SendVob("install.stage.error", imgprofile.name, e)
         raise
      finally:
         self._freeLock()

   def Remediate(self, checkmaintmode = True):
      """Remediate applies a previously staged image.  The exact
         behavior depends on the *Installer subclasses carrying out
         the work. Maintenance Mode is checked if any VIBs require it.
         The databases will be locked for writing; therefore only one
         process may invoke Remediate() at a time.
         Stage must be called before Remediate().
         If not, Remediate may not do anything.

         Parameters:
            * checkmaintmode - Boolean, check that the ESX host is in
                               maintenance mode if the VIBs require it.
                               Note that for bootbank-only installs,
                               Remediate is a NOP, so maintenance mode is
                               not required.

         Returns:
            None

         Exceptions:
            InstallationError
            LockingError      - Unable to lock database for writing. Most
                                likely another thread has invoked Remediate().
            MaintenanceModeError - Maintenance mode was required but the host
                                is not in maintenance mode.
            HostNotChanged    - If host is not changed.
            NeedsRebootResult - Remediation is bootbank only install, reboot is
                                required.
      """
      self._getLock()
      finishedinstallers = set()
      try:
         for installer in self._GetOrderedInstallers():
            try:
               if 'live' in finishedinstallers \
                     and installer.installertype == 'boot':
                  installer.CacheNewImage(self.installers['live'])
               else:
                  installer.Remediate(checkmaintmode)
            except Errors.HostNotChanged, e:
               continue

            finishedinstallers.add(installer.installertype)
      except Errors.NormalExit:
         raise
      except Exception, e:
         msg = "Could not install image profile: %s" % (e)
         # See #bora/apps/addvob/addvob.c for the vob format string.
         self.SendVob("install.error", e)
         raise
      finally:
         self._freeLock()

      log.debug('Host is remediated by installer: %s' % (
         ', '.join(finishedinstallers)))
      if  len(finishedinstallers) == 0:
         raise Errors.HostNotChanged('Host is not changed.')
      elif 'live' not in finishedinstallers \
            and 'boot' in finishedinstallers:
         raise Errors.NeedsRebootResult()
      elif 'live' in finishedinstallers \
            and 'boot' not in finishedinstallers:
         raise Errors.LiveInstallOnlyResult()

   @staticmethod
   def SendVob(eventid, *args):
      """Sends a VOB (VmKernel Observation) event.   VOB events allow important
         host events to be easily tracked in various clients.  This method
         is designed to send VOBs related to esximage.

         Parameters:
            * eventid - The VOB ID relative to vob.user.esximage, and
              (more importantly) the event type ID relative to
              esx.problem.esximage of the event that will be posted to
              hostd.  That "esx.problem.esximage." + eventid string
              will then be associated with a KB article.
            * args - The arguments associated with the vob.  See
              #bora/apps/addvob/addvob.c for details.
      """
      vob = vmware.vob.createVob("vob.user.esximage." + eventid, *args)
      try:
         vob.send()
      except Exception, e:
         # Not logging the arguments, as syslog has length limits.
         log.error('Failed to send vob %s: %s' % (eventid, e))

   @staticmethod
   def SendConsoleMsg(msg):
      """Sends msg to console. The message will be displayed in red on
         informational console to warn user about potential issues.

         Parameters:
            * msg  - A string describing the warning.
      """
      if HAVE_SPICY:
         spicy.set_sysalert(msg)

   @staticmethod
   def _getHostAcceptance():
      #
      # If there is an error or the configured value does not make sense,
      # we return a default.
      try:
         res, out = runcommand.runcommand([ADVCFG, '-U',
                                           'host-acceptance-level', '-G'])
      except runcommand.RunCommandError, e:
         log.error('Unable to invoke %s' % (ADVCFG))
         errMsg = 'Unable to invoke %s' % (ADVCFG)
         raise Errors.AcceptanceGetError(errMsg)

      hostaccept = out.strip()
      if hostaccept in Vib.ArFileVib.ACCEPTANCE_LEVELS:
         return hostaccept
      elif hostaccept:
         log.error("Illegal acceptance level '%s' obtained from "
                   "configuration" % (hostaccept))
         return hostaccept
      else:
         log.error("No host acceptance level is configured")
         return ""

   @staticmethod
   def _setHostAcceptance(acceptance):
      assert acceptance in Vib.ArFileVib.ACCEPTANCE_LEVELS
      try:
         runcommand.runcommand([ADVCFG, '-U', 'host-acceptance-level',
                                '-S', acceptance], returnoutput=False)
      except runcommand.RunCommandError, e:
         raise Errors.AcceptanceConfigError("Unable to execute %s: %s" %
                                            (ADVCFG, str(e)))

   def _getLock(self):
      try:
         self._lock.Lock()
      except LockFile.LockFileError, e:
         log.warning('Failed to acquire lock: %s' % (str(e)))
         raise Errors.LockingError("Another process is updating the ESX "
                                   "image. Please try again later.")

   def _freeLock(self):
      try:
         self._lock.Unlock()
      except Exception, e:
         raise Errors.LockingError("Unable to free lock: %s" % (str(e)))
