#!/usr/bin/python
########################################################################
# Copyright (C) 2011 VMWare, Inc.
# All Rights Reserved
########################################################################
#
# This module implements some work-arounds for broken HTTPS proxy support in
# the httplib and urllib2 modules included with the Python distribution.
# See http://bugs.python.org/issue1424152 for more information.
#
# Additionally, urllib.getproxies() is broken on Windows:
# 1. If the 'Use the same proxy server for all protocols' option is checked in
#    Internet Options -> Local Area Network (LAN) Settings -> Proxy Settings,
#    then proxies are returned for 'ftp' and 'http' protocols, but not 'https'.
#    Maybe the author at the time realized that HTTPS proxy support was
#    broken, and so didn't bother? Who knows.
# 2. If the 'Use the same proxy server for all protocols' option is not
#    checked, then urllib.getproxies() will return a proxy for the 'https'
#    protocol (assuming one was configured). However, it will return the proxy
#    URL as https://proxy.vmware.com:3128 when it should be
#    http://proxy.vmware.com:3128.
#
# We try to fix these issues by re-defining urllib.getproxies_registry() on
# Windows platforms.
#
# This code has a lot of dependencies on implementation-specific things in
# httplib, urllib and urllib2, and is likely to break if/when they are updated.
# It seems to work, however, with Python versions 2.4 - 2.6.

import os, socket, httplib, urllib, urllib2

if os.name == 'nt':
   import re
   def getproxies_registry():
      # Copied from urllib.py in Python 2.6.1.
      proxies = {}
      try:
         import _winreg
      except ImportError:
         return proxies
      try:
         internetSettings = _winreg.OpenKey(_winreg.HKEY_CURRENT_USER,
            r'Software\Microsoft\Windows\CurrentVersion\Internet Settings')
         proxyEnable = _winreg.QueryValueEx(internetSettings, 'ProxyEnable')[0]
         if proxyEnable:
            proxyServer = str(_winreg.QueryValueEx(internetSettings,
                                                   'ProxyServer')[0])
            if '=' in proxyServer:
               for p in proxyServer.split(';'):
                  protocol, address = p.split('=', 1)
                  if not re.match('^([^/:]+)://', address):
                     # This line is one fix:
                     address = 'http://%s' % address
                  proxies[protocol] = address
            else:
               # These 4 lines are the other fix:
               if not re.match('^([^/:]+)://', proxyServer):
                  proxyServer = 'http://%s' % proxyServer
               for key in ('http', 'ftp', 'https'):
                  proxies[key] = proxyServer
         internetSettings.Close()
      except (WindowsError, ValueError, TypeError):
         pass
      return proxies
   # Note: This will replace the function for everyone who imports urllib, not
   #       just this module.
   urllib.getproxies_registry = getproxies_registry

class ConnectProxyError(httplib.error):
   pass

class HTTPSProxyConnection(httplib.HTTPSConnection):
   '''Like httplib.HTTPSConnection, but we establish a tunnel through a proxy
      first.
   '''
   def __init__(self, host, port=None, key_file=None, cert_file=None,
                strict=None, proxy=None):
      httplib.HTTPSConnection.__init__(self, host, port, key_file, cert_file,
                                       strict)
      if proxy and proxy.count('://'):
         proto, proxy = proxy.split('://', 1)
         if proto != 'http':
            raise ValueError("Unknown proxy protocol '%s'" % proto)
      self.proxy = proxy

   def connect(self):
      if not self.proxy:
         return httplib.HTTPSConnection.connect(self)
      # Let's talk to our proxy server:
      h = httplib.HTTPConnection(self.proxy)
      h.request('CONNECT', '%s:%d' % (self.host, self.port))
      resp = h.getresponse()
      if resp.status != 200:
         raise ConnectProxyError("Error connecting to proxy: %s %s" %
                                 (resp.status, resp.reason))
      # If our proxy server gave us a green light, anything we send/receive
      # on this socket from here on out should be tunneled to the remote
      # server.  Re-use the socket (resp.fp) from our connection to the proxy,
      # and set up our SSL connection now.
      try:
         import ssl
         self.sock = ssl.wrap_socket(resp.fp, self.key_file, self.cert_file)
      except ImportError:
         sslsock = socket.ssl(resp.fp, self.key_file, self.cert_file)
         self.sock = httplib.FakeSocket(resp.fp, sslsock)

class ProxyHandler(urllib2.ProxyHandler, urllib2.AbstractHTTPHandler):
   '''Sub-classes urllib2.ProxyHandler in order to handle https:// connections
      using the HTTPSProxyConnection class.
   '''
   def __init__(self, *args, **kwargs):
      urllib2.AbstractHTTPHandler.__init__(self)
      urllib2.ProxyHandler.__init__(self, *args, **kwargs)
      # See the call to setattr in urllib2.ProxyHandler.__init__ to understand
      # why this is (unfortunately) necessary.
      if hasattr(self, 'https_open') and hasattr(httplib, 'HTTPS'):
         self.https_open = self._https_open

   https_request = urllib2.AbstractHTTPHandler.do_request_

   def _https_open(self, req):
      # This is essentially duplicated from
      # urllib2.AbstractHTTPHandler.do_open.  :-(
      host = req.get_host()
      if not host:
         raise urllib2.URLError('no host given')

      # Here is our change to use HTTPSProxyConnection if we have an https
      # proxy.
      if self.proxies.has_key('https'):
         h = HTTPSProxyConnection(host, proxy=self.proxies['https'])
      else:
         h = httplib.HTTPSConnection(host)
      # End our changes.

      h.set_debuglevel(self._debuglevel)

      headers = dict(req.headers)
      headers.update(req.unredirected_hdrs)
      headers["Connection"] = "close"
      headers = dict(
         (name.title(), val) for name, val in headers.items())
      try:
         h.request(req.get_method(), req.get_selector(), req.data, headers)
         r = h.getresponse()
      except socket.error, err:
         raise urllib2.URLError(err)

      r.recv = r.read
      fp = socket._fileobject(r)

      resp = urllib.addinfourl(fp, r.msg, req.get_full_url())
      resp.code = r.status
      resp.msg = r.reason
      return resp

if __name__ == '__main__':
   import sys, optparse

   parser = optparse.OptionParser()
   parser.set_usage('%prog [options] URL')
   parser.add_option('-p', '--proxy', dest='proxy',
                     metavar='proto://host:port')
   options, args = parser.parse_args()
   if len(args) < 1:
      parser.error('URL required')
   if len(args) > 1:
      parser.error('Too many arguments')

   proxies = dict()
   if options.proxy:
      proxies['ftp'] = options.proxy
      proxies['http'] = options.proxy
      proxies['https'] = options.proxy
   opener = urllib2.build_opener(ProxyHandler(proxies))
   f = opener.open(args[0])
   sys.stdout.write(f.read())
