#!/usr/bin/python

import errno
import fcntl
import logging
import os
import select
import signal
import subprocess
import sys

"""This module provides the runcommand function, a convenience function to
   run a command and collect the output.
"""

log = logging.getLogger(__name__)

MAXOUTPUTSIZE = 1048576 # 1 MB
ONVISOR = os.uname()[0].lower().startswith('vmkernel')
try:
   import vmkuser
   HAVEVMKUSER = True
except:
   HAVEVMKUSER = False

class RunCommandError(Exception):
   pass

class VisorPopen(object):
   """A rather simple-minded emulation of the Popen class for use on Visor."""

   def __init__(self, args):
      """Class constructor.
            Parameters:
               * args - Either a sequence or string. If a string, the string
                        will be executed via a shell. If a sequence, the first
                        item is the program to execute, and the remaining items
                        are arguments to the program.
         Returns: A VisorPopen object.
         Raises:
            * RuntimeError - If the vmkuser module is not available.
            * OSError      - If the command cannot be executed.
      """
      if not HAVEVMKUSER:
         msg = "Class %s requires vmkuser module" % self.__class__.__name__
         raise RuntimeError(msg)

      if isinstance(args, basestring):
         args = ['/bin/sh', '-c', args]
      else:
         args = list(args)

      self.returncode = None

      fromchildfd, toparentfd = os.pipe()
      self.stdout = os.fdopen(fromchildfd)
      initfds = [sys.stdin.fileno(), toparentfd, toparentfd]
      self.pid = vmkuser.ForkExec(args[0], args, None, -1, initfds, -1, -1, 0)
      os.close(toparentfd)

   # The rest of this is shamelessly borrowed from Python's subprocess module.
   def _handle_exitstatus(self, sts):
      if os.WIFSIGNALED(sts):
         self.returncode = -os.WTERMSIG(sts)
      elif os.WIFEXITED(sts):
         self.returncode = os.WEXITSTATUS(sts)
      else:
         # Should never happen
         raise RuntimeError("Unknown child exit status!")

   def poll(self):
      if self.returncode == None:
         try:
            pid, sts = os.waitpid(self.pid, os.WNOHANG)
            if pid == self.pid:
               self._handle_exitstatus(sts)
         except os.error:
            pass
      return self.returncode

def terminateprocess(p):
   # Get progressively heavy-handed.
   for s in (signal.SIGINT, signal.SIGTERM, signal.SIGKILL):
      try:
         os.kill(p.pid, s)
      except OSError, e:
         # Maybe the process already exited. We could check for ESRCH, but we
         # wouldn't really do anything differently, anyway.
         return
      timeout = 3.0
      while timeout > 0.0:
         # Use select as a per-thread 'sleep'.
         select.select([], [], [], 0.02)
         timeout -= 0.02
         if p.poll() != None:
            return

def runcommand(args, outfile=None, returnoutput=True, timeout=0.0):
   """Execute a command, optionally collecting output.
         Parameters:
            * args         - May be either a sequence of arguments, or a
                             string. If a string, the string will be passed to
                             a shell for execution. If a sequence of arguments,
                             the first item is the program to execute, and the
                             remaining items are arguments to the program. Note
                             that when a sequence of arguments is specified,
                             the shell is not invoked, so arguments are not
                             subject to shell interpolation.
            * outfile      - If specified, may be either a file-like object
                             with a write method, or None. If a file-like
                             object, command  output will be written using its
                             write() method.
            * returnoutput - If True, output will be returned by the function.
                             Note that, when True, an exception will be raised
                             if output exceeds 1 MB. Defaults to True.
            * timeout      - A float value specifying the amount of time,
                             in seconds, to wait for output or completion. If
                             the timeout expires before output is read or the
                             process terminates, an exception is raised. A
                             value of 0 disables the timeout, and this function
                             may block indefinitely. Defaults to 0.
         Returns: A tuple of the command's exit value and output.
         Raises:
            * RunCommandError - All errors are caught and raised as this
                                exception. May happen if Popen fails (e.g.
                                command or process image does not exist or
                                cannot be executed), if there is an I/O error
                                reading output, or if the output exceeds 1024
                                characters when returnoutput is True.
   """

   log.info("runcommand called with: args = '%s', outfile = '%s', "
            "returnoutput = '%s', timeout = '%s'." %
            (args, outfile, returnoutput, timeout))

   try:
      if ONVISOR and HAVEVMKUSER:
         p = VisorPopen(args)
      elif isinstance(args, basestring):
         p = subprocess.Popen(args, stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT, shell=True)
      else:
         p = subprocess.Popen(args, stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT)
   except Exception, e:
      raise RunCommandError("Error running command '%s': %s" % (args, e))

   # Set O_NONBLOCK on pipe, else we can hang on a read. See PR 110272.
   flags = fcntl.fcntl(p.stdout, fcntl.F_GETFL) | os.O_NONBLOCK
   fcntl.fcntl(p.stdout, fcntl.F_SETFL, flags)

   pollobj = select.poll()
   pollobj.register(p.stdout, select.POLLIN|select.POLLPRI)

   data = ''
   out = ''
   while p.poll() == None:
      try:
         if timeout > 0:
            ready = pollobj.poll(timeout * 1000)
         else:
            # See the "baddaemon" test in test/test_runcommand.py. By using a
            # timeout here, it gives us a chance to re-evaluate the conditional
            # in the while loop. This will catch daemon processes which don't
            # close their stdin/stdout/stderr.
            ready = pollobj.poll(200)
      except select.error, (n, strerror):
         # If we're interrupted by a signal, that's ok. Otherwise, it's
         # something bad and we should bail.
         if n == errno.EINTR:
            continue
         msg = ("Error waiting for output from command '%s': %s" %
                (args, strerror))
         raise RunCommandError(msg)
      if timeout and not ready and p.poll() == None:
         # The select() call returned before we had a signal or output, and the
         # command hasn't finished yet.
         terminateprocess(p)
         msg = ("Timeout (%d seconds) expired waiting for output from "
                "command '%s'." % (timeout, args))
         raise RunCommandError(msg)

      try:
         # We read up to 1024 bytes at a time, to avoid usurping lots of
         # memory. If there is more data in the pipe, the next select will just
         # return right away.
         data = p.stdout.read(1024)
      except IOError, e:
         if e.errno == errno.EAGAIN:
            # We've set O_NONBLOCK, so the read() may return EAGAIN.
            continue
         else:
            msg = "Error reading output from command '%s': %s" % (args, e)
            raise RunCommandError(msg)

      # If we've read data, send it to outfile and/or add it to the collected
      # data to return.
      if outfile:
         outfile.write(data)
      if returnoutput:
         if len(out) + len(data) > MAXOUTPUTSIZE:
            msg = "Output from command '%s' is too large." % (args,)
            raise RunCommandError(msg)
         out += data
      data = ''

   # Read in any more data in the pipe.
   try:
      data = p.stdout.read(1024)
   except IOError:
      data = ''
   while data:
      if outfile:
         outfile.write(data)
      if returnoutput:
         if len(out) + len(data) > MAXOUTPUTSIZE:
            msg = "Output from command '%s' is too large." % (args,)
            raise RunCommandError(msg)
         out += data
      try:
         data = p.stdout.read(1024)
      except IOError:
         data = ''

   # Get the status of the child.
   rc = p.poll()

   # Note: subprocess.Popen's poll() method will return a negative value if the
   #       process exited due to a signal.
   if rc >= 0:
      return rc, out
   else:
      # If we exited due to a signal, raise an exception.
      msg = "Command '%s' terminated due to signal %d" % (args, -rc)
      raise RunCommandError(msg)
